#include "main.h"


/*--------------------------------------------------------
    function: init_render_engine
    purpose:		performs all openGL state initializing. 
  -------------------------------------------------------*/
void init_render_engine(void)
{
    cerr << "Init:\t\tinit_render_engine()";
    
    glClearColor( 0.0, 0.0, 0.0, 1.0 );
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);  
    glLoadIdentity();

    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);

    glEnable(GL_DEPTH_TEST);
    
    glDisable(GL_TEXTURE_2D);
    
    glShadeModel(GL_FLAT);
    
    if (image_quality == HIGH_QUALITY)
    {
	    glHint( GL_POLYGON_SMOOTH_HINT, GL_NICEST);
	    glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	    /*
	    glEnable(GL_FOG);
	    glFogi(GL_FOG_MODE, GL_LINEAR);
	    glFogfv(GL_FOG_COLOR, mat_grey);
	    glFogf(GL_FOG_DENSITY, 1.0);
	    glFogf(GL_FOG_START, 0.1);
	    glFogf(GL_FOG_END, 40.0);
	    glHint(GL_FOG_HINT, GL_NICEST);
	    */
    }

    glEnable(GL_LIGHTING);

    glEnable(GL_LIGHT0);
    glEnable(GL_LIGHT1);
    glEnable(GL_LIGHT2);    
    
    glEnable(GL_LIGHT3);
    glEnable(GL_LIGHT4);
    glEnable(GL_LIGHT5);

    glClearColor( 0.0, 0.0, 0.0, 1.0 );
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    gluPerspective( 60, ASPECT_RATIO, 0.15, 100 );
    
    cerr << "\tdone." << endl;
    return;
}

/*--------------------------------------------------------
    function: init_demo
    purpose:		initializes objects for the demo section,
              here 60 cubes.
  -------------------------------------------------------*/
void init_demo(void)
{
    cerr << "Init:\t\tinit_demo()";

    /* les supers cubes qui tournent */
    for( int i = 0 ; i < 60 ; i++)
    {
	    object_list.addItem(new Actor());
	    
	    object_list[i]->setPosition(-3.1, 1.6, -6.0);
	    object_list[i]->setGround(-3.0);
	    object_list[i]->setSize(0.5, 0.5, 0.5);
	    object_list[i]->setBoundCheck(false);
	    object_list[i]->setSpeed( 0.0, 0.0, 0.0);
	    
	    if (i < 20) {
		    object_list[i]->setColor( 1.0, 0.8, 0.5, 1.0);
		    object_list[i]->setPivot( -3, 0.0, 0.0);
		    object_list[i]->setRotationSpeed( 0.8, -1.0, i*0.1);
	    } else {
		    if (i < 40) {
			    object_list[i]->setColor( 0.8, 1.0, 0.5, 1.0);
			    object_list[i]->setPivot( -4.5, 0.0, 0.0);
			    object_list[i]->setRotationSpeed( 0.8, 1.0, (i-20)*0.1);
			} else {
			    object_list[i]->setColor( 0.8, 0.5, 1.0, 1.0);	    
			    object_list[i]->setPivot( -1.5, 0.0, 0.0);
				object_list[i]->setRotationSpeed( 0.8, 1.0, (i-40)*0.1);
			}
		}
		
	    object_list[i]->setMesh(SOLID_CUBE); 
	    object_list[i]->setElasticity(0.8);
	    object_list[i]->setUpperLimit(-200, -200, -200);
	    object_list[i]->setDownLimit(200, 200, 200);
    }
    cerr << "\t\tdone." << endl;
    glDisable(GL_TEXTURE_2D);
    
}    

/*--------------------------------------------------------
    function: init_game
    purpose:		initialise la partie jeu, les 3 boules et la
		      queue. Creation des display-lists.
  -------------------------------------------------------*/
void init_game(void)
{
    cerr << "Init:\t\tinit_game()";
    object_list.deleteList();
		
	/* les boules */
    for(int i = 0 ; i < 3 ; i++)
    {
		object_list.addItem(new Actor());
		object_list[i]->setPivot( 0.0, 0.0, 0.0);
		object_list[i]->setGround(-0.9);
		object_list[i]->setSize(0.3, 0.3, 0.3);
		object_list[i]->setRotationSpeed(0.0, 0.0, 0.0);
	    object_list[i]->setMesh(SOLID_SPHERE);
		object_list[i]->setSpeed( 0.0, 0.0, 0.0);
		
		if ( i == 0 ) // la blanche
		{
		    object_list[i]->setColor(1.0, 1.0, 1.0, 1.0);
		    object_list[i]->setPosition( 0.0f, -0.9f, -3.5f);
		}
		    
		if ( i == 1 ) //la rouge 
		{
		    object_list[i]->setColor(1.0, 0.0, 0.0, 1.0);
		    object_list[i]->setPosition( 1.0f, -0.9f, 1.5f);
		}
		    
		if ( i == 2 ) //la bleu
		{
		    object_list[i]->setColor(0.0, 0.0, 1.0, 1.0);
		    object_list[i]->setPosition( -0.0f, -0.9f, 1.5f);
		}
		
		object_list[i]->setElasticity(0.8);
	    object_list[i]->setUpperLimit(-2.45, -100.0, -5.0);
		object_list[i]->setDownLimit( 2.0, 100.0, 1.9);
	    object_list[i]->setMotion( 0.0, 0.0, 0.0 );
	    
#ifdef _USE_3DFX_
	    object_list[i]->setFriction( 0.002); // for hardware rendering
#else
	    object_list[i]->setFriction( 0.005);	// for software rendering
#endif	    
    }
    object_count += 3;
        
    /* la queue */
    object_list.addItem(new Actor());
    
    object_list[QUEUE]->setPosition(0.0, -0.9, 0);
    object_list[QUEUE]->setPivot(object_list[0]->getPosition());
    
    object_list[QUEUE]->setGround(-0.9);
    object_list[QUEUE]->setSize(0.05, 0.05, 2.0);
    object_list[QUEUE]->setRotation(-6.0, 180.0, 0.0);
    
    object_list[QUEUE]->setRotationSpeed(0.0, 0.0, 0.0);
    object_list[QUEUE]->setMesh(SOLID_CUBE);
    object_list[QUEUE]->setColor( 0.5, 0.1, 0.1, 1.0);
    object_list[QUEUE]->setElasticity(0.8);
    object_list[QUEUE]->setUpperLimit(-200,-200,-200);
    object_list[QUEUE]->setDownLimit(200,200,200);
    object_list[QUEUE]->setMotion(0.0, 0.0, 0.0);
    object_count++;


	/* les ombres */
    for(int i = object_count ; i < object_count+3 ; i++)
    {
		object_list.addItem(new Actor());
		object_list[i]->setPivot( 0.0, 0.0, 0.0);
		object_list[i]->setGround(-1.0);
		object_list[i]->setSize(0.3, 0.01, 0.3);
		object_list[i]->setRotationSpeed(0.0, 0.0, 0.0);
	    object_list[i]->setMesh(SOLID_SPHERE);
		object_list[i]->setSpeed( 0.0, 0.0, 0.0);
		
	    object_list[i]->setColor(0.0, 0.0, 0.0, 1.0);
		
		if ( i == OMBRE_BLANCHE ) // la blanche
		    object_list[i]->setPosition( 0.0f, -1.0f, -3.5f);
		    
		if ( i == OMBRE_ROUGE ) //la rouge 
		    object_list[i]->setPosition( 1.0f, -1.0f, 1.5f);
		    
		if ( i == OMBRE_BLEU ) //la bleu
		    object_list[i]->setPosition( -0.0f, -1.0f, 1.5f);
		
		object_list[i]->setElasticity(0.8);
	    object_list[i]->setUpperLimit(-2.45, -100.0, -5.0);
		object_list[i]->setDownLimit( 2.0, 100.0, 1.9);
	    object_list[i]->setMotion( 0.0, 0.0, 0.0 );
	    
#ifdef _USE_3DFX_
	    object_list[i]->setFriction( 0.002); // for hardware rendering
#else
	    object_list[i]->setFriction( 0.005);	// for software rendering
#endif	    
    }
    object_count += 3;

    // builds all display_lists
    make_room();
    make_billard();
    make_lightmap();
    make_bank();
    
    if (image_quality == LOW_QUALITY)
	    glShadeModel(GL_FLAT);
	 else
	    glShadeModel(GL_SMOOTH); 
	    
    if (image_quality == HIGH_QUALITY)
	    load_objects();

    eye.setPosition( -3.0, 1.5, 3.0);
    retarget_queue(players_turn);
    
    cerr << "\t\tdone."<<endl;
    game_mode = 1;
}

/*--------------------------------------------------------
    function: load_textures
    purpose:		charge les textures dans la liste chaine de
              textures.
  -------------------------------------------------------*/
void load_textures(void)
{
    cerr << "Init:\t\tload_textures()";

/* TEX_GROUND */
    map_list.addItem(new Texture( GIF, "./matlib/grounds/parquet.gif",128, 1) );
    
/* TEX_WALL_1 */
    map_list.addItem(new Texture( PNG, "./matlib/walls/concwhit.png",128, 1) ); 
    
/* TEX_WALL_2 */
    map_list.addItem(new Texture( GIF, "./matlib/woods/bubbing.gif",128, 1) );
	
/* TEX_ROOF */
    map_list.addItem(new Texture( PNG, "./matlib/walls/concwhit.png",128, 1) );
    
/* TEX_SPOTLIGHT */
    map_list.addItem(new Texture( PNG, "./matlib/lights/light.png",128, 1) );
    
/* TEX_PAINTING */
    map_list.addItem(new Texture( PNG, "./matlib/walls/pool.png",128, 1) );
    
/* TEX_INTRO */
    map_list.addItem(new Texture( PNG, "./matlib/intro.png",255, 1) );
    
/* TEX_LIGHTMAP */
    map_list.addItem(new Texture( PNG, "./matlib/lights/lightmap2.png",255, 0) );
    
/* TEX_SHADOWMAP */
    map_list.addItem(new Texture( PNG, "./matlib/walls/barber.png",128, 1) );
    
    for(int i = 0 ; i < map_list.getLength(); i++)
	    map_list[i]->Disable();
	    
    cerr << "\t\tdone." << endl;
}

/*--------------------------------------------------------
    function: load_objects
    purpose:		charge les objets dans la liste chaine des
              acteurs.
  -------------------------------------------------------*/
void load_objects(void)
{
    for(int i = object_count; i < object_count + 2 ; i++)
    {
	    object_list.addItem(new Actor());
	    object_list[i]->setPosition(0.0, 0.0, 0.0);
	    object_list[i]->setPivot(0.0, 0.0, 0.0);
	    object_list[i]->setGround(0.0);
	    object_list[i]->setSize(1.0, 1.0, 1.0);
      	
	    object_list[i]->setRotationSpeed(0.0, 0.0, 0.0);
	  
	    object_list[i]->setMesh(CUSTOM_MESH); 
	    object_list[i]->setColor(0.0, 0.0, 0.0, 1.0);
	    object_list[i]->setElasticity(0.8);
      
	    object_list[i]->setMotion( 0.0, 0.0, 0.0);
	    
	    object_list[i]->setUpperLimit(-200,-200,-200);
	    object_list[i]->setDownLimit(200,200,200);
    }

    object_list[object_count]->loadFromFile("./meshes/teapot.msh");	// la theiere
    object_list[object_count]->setPosition(7.0, -0.5, -3.5);
    object_list[object_count]->setColor(1.0, 1.0, 0.0, 1.0);    
    object_list[object_count]->setRotation(-90.0, 0.0, 123 );
    object_list[object_count]->scaleMesh(0.5);
    object_list[object_count]->makeList();
    
    object_list[object_count+1]->loadFromFile("./meshes/table.msh"); // la table
    object_list[object_count+1]->setColor(0.4, 0.1, 0.1, 1.0);    
    object_list[object_count+1]->setPosition(7.0, -3.0, -3.5);
    object_list[object_count+1]->setRotation(-90.0, 0.0, 0.0 );
    object_list[object_count+1]->scaleMesh(1.5);
    object_list[object_count+1]->makeList();
    
    object_count += 2;
}


/*--------------------------------------------------------
    function: retarget_queue
    purpose:		set the queue position around the player's
		      ball
    parameter: the ball of the player
  -------------------------------------------------------*/
void retarget_queue(int which_ball)
{
    vertex v,p;
    v = object_list[which_ball]->getPosition() ;
    p = object_list[which_ball]->getPivot() ;
				    
    object_list[QUEUE]->setPosition(v.x ,-0.9, v.z);
    object_list[QUEUE]->setPivot(0.0, 0.0, 1.0);
}

/*--------------------------------------------------------
    function: hit_ball
    purpose:		calcule la force et l'angle de frappe de la
              queue sur la boule.
  -------------------------------------------------------*/
void hit_ball(int which_ball)
{
    float sx,sz;
    float r;
    vertex cr;
    
    r = PI_2/360.0 ;
    
    cr = object_list[QUEUE]->getRotation();
    cr.y *= -1 ;
    
    sx = ( sin(cr.y*r) * force) ;
    sz = ( -cos(cr.y*r)* force) ;
    
    object_list[which_ball]->setMotion( sx, 0.0, sz);
    
    /*- et aussi l'ombre -*/
    object_list[which_ball+4]->setMotion( sx, 0.0, sz);    
}

/*--------------------------------------------------------
    function: impact
    purpose:		calcule les nouvelles trajectoires des 2 
              boules en collision.
  -------------------------------------------------------*/
void impact(int obj_1, int obj_2)
{
    register float a, b, a_2, b_2, ab, d;

    vertex v1, v2 ;
    vertex r1, r2;

	v1 = object_list[obj_1]->getPosition();
	v2 = object_list[obj_2]->getPosition();
	
    a = v1.x - v2.x ;
    b = v1.z - v2.z ;
	
	v1 = object_list[obj_1]->getSpeed() ;
	v2 = object_list[obj_2]->getSpeed() ;
	
    // some precalculation to speed things up
    a_2 = a * a ;
    b_2 = b * b ;
    
    ab = a * b ;
    d = a_2 + b_2 ;

    r1.x = (v2.z*ab - v1.z*ab + b_2*v1.x + a_2*v2.x) / d;
    r1.y = 0.0 ;
    r1.z = (v2.z*b_2 + v2.x*ab - v1.x*ab + a_2*v1.z) / d;
    
    r2.x = (a_2*v1.x - v2.z*ab + v1.z*ab + b_2*v2.x )/ d ;
    r2.y = 0.0 ;
    r2.z = (v1.x*ab + v1.z*b_2 - v2.x*ab + a_2*v2.z) / d ;

    object_list[obj_1]->setSpeed(r1);
    object_list[obj_2]->setSpeed(r2);

    /*- deplacer les ombres -*/
    object_list[obj_1+4]->setSpeed(r1);
    object_list[obj_2+4]->setSpeed(r2);
}

/*--------------------------------------------------------
    function: check_collision
    purpose:		effectue les testes de collisions afin de
			  determiner quelle boules sont en contact.
  -------------------------------------------------------*/
void check_collision(void)
{
    float r = 0.2; // radius of the sphere
    
    vertex b1, b2, b3 ;
    
    b1 = object_list[0]->getPosition();
	b2 = object_list[1]->getPosition();
	b3 = object_list[2]->getPosition();	
		

    if (( b1.x + r > b2.x )&&( b1.x - r < b2.x))
	    if (( b1.z + r > b2.z )&&( b1.z - r < b2.z))
		    impact(0,1);
	    
    if (( b1.x + r > b3.x )&&( b1.x - r < b3.x))
	    if (( b1.z + r > b3.z )&&( b1.z - r < b3.z))
		    impact(0,2);

    if (( b2.x + r > b3.x )&&( b2.x - r < b3.x))
		if (( b2.z + r > b3.z )&&( b2.z - r < b3.z))
		    impact(1,2);
}

/*--------------------------------------------------------
    function: set_camera
    purpose:		effectue toutes les transformations de camera
              et prepare la MODELVIEW_MATRIX.
		      Active la lumiere frontale et la barre de 
			  force.
  -------------------------------------------------------*/
void set_camera()
{
    vertex cr,cp;  // camera rotation, camera position 
    
    eye.Run();
    
    cr = eye.getRotation();
    cp = eye.getPosition();
    
    glLoadIdentity();
    gluPerspective( 60, ASPECT_RATIO, 0.15, 100 );
    
    // la barre de force
    glPushMatrix();
	    glDisable(GL_LIGHTING);
	    glColor3f( 0.0, 0.5, 1.0);
	    glTranslatef(0.7, -0.5 + force/2, -1.0);
	    glScalef(0.05, force, 0.05);
	    glutSolidCube(1);
	    glEnable(GL_LIGHTING);
    glPopMatrix();
    
	// light 3 lumiere frontale
    glPushMatrix();
        glLightfv(GL_LIGHT3, GL_POSITION, light_pos3);
	    glLightfv(GL_LIGHT3, GL_AMBIENT, mat_amb);
	    glLightfv(GL_LIGHT3, GL_DIFFUSE, mat_white);
        
        glLightf(GL_LIGHT3, GL_CONSTANT_ATTENUATION, 1.0);
	    glLightf(GL_LIGHT3, GL_LINEAR_ATTENUATION, 2.5);
        glLightf(GL_LIGHT3, GL_QUADRATIC_ATTENUATION, 2.5);
    glPopMatrix();
    
    glRotatef(cr.y, 1.0, 0.0, 0.0);
    glRotatef(cr.x, 0.0, 1.0, 0.0);    
    glTranslatef( -cp.x, -cp.y, -cp.z);
}

/*--------------------------------------------------------
    function: set_lighting
    purpose:		active les lumires.
  -------------------------------------------------------*/
void set_lighting(void)
{
    vertex v; 
    
    eye.getPosition();
    cam[0] = v.x; cam[1] = v.y; cam[2] = v.z;
    glLightModelfv(GL_LIGHT_MODEL_LOCAL_VIEWER, cam);
  
    glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);	
    
    // LIGHT 0
    glPushMatrix();
        glLightfv(GL_LIGHT0, GL_POSITION, omni_light_pos);
        glLightfv(GL_LIGHT0, GL_AMBIENT, mat_amb);
        glLightfv(GL_LIGHT0, GL_DIFFUSE, mat_white);
        glLightfv(GL_LIGHT0, GL_SPECULAR, mat_spec);
        
        glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.0);
	    glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 2.5);
        glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 2.5);
    glPopMatrix();

    // LIGHT 1
    glPushMatrix();
        glLightfv(GL_LIGHT1, GL_POSITION, light_pos1);
        glLightfv(GL_LIGHT1, GL_AMBIENT, mat_amb);
        glLightfv(GL_LIGHT1, GL_DIFFUSE, mat_white);
        
        glLightf(GL_LIGHT1, GL_CONSTANT_ATTENUATION, 1.0);
	    glLightf(GL_LIGHT1, GL_LINEAR_ATTENUATION, 2.5);
        glLightf(GL_LIGHT1, GL_QUADRATIC_ATTENUATION, 2.5);
    glPopMatrix();

    // light 2
    glPushMatrix();
        glLightfv(GL_LIGHT2, GL_POSITION, light_pos2);
	    glLightfv(GL_LIGHT2, GL_AMBIENT, mat_amb);
	    glLightfv(GL_LIGHT2, GL_DIFFUSE, mat_white);

        glLightf(GL_LIGHT2, GL_CONSTANT_ATTENUATION, 1.0);
	    glLightf(GL_LIGHT2, GL_LINEAR_ATTENUATION, 2.5);
        glLightf(GL_LIGHT2, GL_QUADRATIC_ATTENUATION, 2.5);
    glPopMatrix();
    
    // light 4
    glPushMatrix();
        glLightfv(GL_LIGHT4, GL_POSITION, light_pos4);
	    glLightfv(GL_LIGHT4, GL_AMBIENT, mat_amb);
	    glLightfv(GL_LIGHT4, GL_DIFFUSE, mat_white);

        glLightf(GL_LIGHT4, GL_CONSTANT_ATTENUATION, 1.0);
	    glLightf(GL_LIGHT4, GL_LINEAR_ATTENUATION, 2.5);
        glLightf(GL_LIGHT4, GL_QUADRATIC_ATTENUATION, 2.5);
    glPopMatrix();
    
	// light 5
    glPushMatrix();
        glLightfv(GL_LIGHT5, GL_POSITION, light_pos5);
	    glLightfv(GL_LIGHT5, GL_AMBIENT, mat_amb);
	    glLightfv(GL_LIGHT5, GL_DIFFUSE, mat_white);

        glLightf(GL_LIGHT5, GL_CONSTANT_ATTENUATION, 1.0);
	    glLightf(GL_LIGHT5, GL_LINEAR_ATTENUATION, 2.5);
        glLightf(GL_LIGHT5, GL_QUADRATIC_ATTENUATION, 2.5);
    glPopMatrix();
}

/*--------------------------------------------------------
    function: run_world
    purpose:		fonction threadable qui va executer les objets
			  de la liste et les placer dans la scne.
  -------------------------------------------------------*/
void run_world(void)
{
    int l,i;
    vertex v,r,p,s; // position, rotation, pivot,size
    l = object_list.getLength();
    
    if (game_mode)
    {    
	    v = object_list[BLANCHE]->getPosition();
	    v.y = -1.0;
	    object_list[OMBRE_BLANCHE]->setPosition(v);
    
	    v = object_list[ROUGE]->getPosition();
	    v.y = -1.0;
	    object_list[OMBRE_ROUGE]->setPosition(v);
    
	    v = object_list[BLEU]->getPosition();
	    v.y = -1.0;
	    object_list[OMBRE_BLEU]->setPosition(v);
    }
    
    for(i = 0 ; i < l ; i++)
    {
        object_list[i]->Run();

	    if (game_mode)
		    check_collision();
	    
	    v = object_list[i]->getPosition();
	    r = object_list[i]->getRotation();
	    p = object_list[i]->getPivot();
	    s = object_list[i]->getSize();

	    glPushMatrix();
            glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, object_list[i]->getColor());
            
            glTranslatef( v.x, v.y, v.z);
            glRotatef( r.y, 0.0, 1.0, 0.0 );
            glRotatef( r.x, 1.0, 0.0, 0.0 );
            glRotatef( r.z, 0.0, 0.0, 1.0 );
            
		    glTranslatef(p.x, p.y, p.z);
		    glScalef(s.x, s.y, s.z);

            switch( object_list[i]->getMesh() )
            {
                case SOLID_SPHERE:
				    glutSolidSphere(s.x, 8, 8);
                    break;
                    
                case SOLID_CUBE:
                    glutSolidCube(1.0);
                    break;

			    case SOLID_CONE:
                    glutSolidCone(1.0, 1.0, 8, 8);
                    break;
		    
			    case CUSTOM_MESH:
				    glCallList(object_list[i]->getListId());
				    break;
             }
        glPopMatrix();
    }
}

/*--------------------------------------------------------
    function: run_demo
    purpose:		fonction principale qui va executer la demo.
  -------------------------------------------------------*/
void run_demo(void)
{
   switch(image_quality)
    {
    	    case LOW_QUALITY:
			glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    
			set_camera();
			set_lighting();    
		    run_world();
		    glutSwapBuffers();
		    frame++;
		    timer += time_slice;
		    break;

	    case NORMAL_QUALITY:
			glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    
			set_camera();
			set_lighting();    
		    run_world();

		    glPushMatrix();
    
			    glEnable(GL_TEXTURE_2D);
			    glTranslatef( -3.0, 4.0, -1.0);
	    
			    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mat_white);
			    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
			    glBindTexture(GL_TEXTURE_2D, map_list[TEX_INTRO]->getRef());
	    
				glScalef( 0.2, 0.2, 0.1);
	    
			    glBegin(GL_QUADS);
					glNormal3f( 0.0, 0.0, 1.0);
            
				    glTexCoord2f(0.0, 0.0);
				    glVertex3f( -4.0, 0.0, 0.0 );
            
				    glTexCoord2f(0.0, 1.0);
					glVertex3f( -4.0, -4.0, 0.0 );
            
				    glTexCoord2f(1.0, 1.0);
				    glVertex3f( 4.0, -4.0, 0.0 );
            
				    glTexCoord2f(1.0, 0.0);
				    glVertex3f( 4.0, 0.0, 0.0 );
			    glEnd();

			    glDisable(GL_TEXTURE_2D);
			
			glPopMatrix();
		    glutSwapBuffers();
		    frame++;
		    timer += time_slice;
		    break;
		    
		    
	    case HIGH_QUALITY:

		    glClear(GL_ACCUM_BUFFER_BIT);
		    
		    for( int i=0 ; i < 4 ; i++ )
		    {
			    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

				set_camera();
			    set_lighting();    
			    run_world();
			    
			    glAccum(GL_ACCUM, 1.0/4.0);
		    }
		    
	        glAccum(GL_RETURN, 1.0);
		    glutSwapBuffers();
		    
		    frame++;
		    timer += time_slice;
		    break;
    }		    
}

/*--------------------------------------------------------
    function: run_game
    purpose:		fonction principale qui va executer le jeu,
	          et appeler les display-lists.
  -------------------------------------------------------*/
void run_game(void)
{
    vertex v;
    
	/* we should wait a bit before putting any GL command after a glClear*/
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );	    
    
    set_camera();
    set_lighting();
    
	v = object_list[players_turn]->getSpeed();    
	
	if ((v.x == 0)&&(v.z == 0)) {
	    retarget_queue(players_turn);    
    }
    
	run_world();
	
    glCallList(TABLE);
    glCallList(ROOM);
    glCallList(COMPTOIR);

    if (image_quality == HIGH_QUALITY)
    {
	    /*  light cones too expensive because of alpha blending   */
	    glPushMatrix();
//		    glDisable(GL_CULL_FACE);
		    glTranslatef(0.26, -3.5, -16.25);
	    
		    for(int i=0 ; i < 7 ; i++)
		    {
			    glPushMatrix();
				    glEnable(GL_BLEND);
				    glDisable(GL_LIGHTING);
				    glColor4f( 0.8, 0.8, 0.2, 0.2 );
				    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			    
					glTranslatef(0.0, 0.0, i*4.0);
				
				    glRotatef(-90, 1.0, 0.0, 0.0);
			    
				    glutSolidCone(2.0, 8.0, 32, 16);
				    glDisable(GL_BLEND);
				    glEnable(GL_LIGHTING);	    
			    glPopMatrix();
		    }
	    glPopMatrix();
//	    glEnable(GL_CULL_FACE);
    }

/* light map test, looks bad 
    glTranslatef(-8.0, 3.0, -13.9);
    glCallList(LIGHT_MAP);
*/    
    glutSwapBuffers();
	frame++;
    timer += time_slice;
}

/*--------------------------------------------------------
    function: switch_demo_to_game
    purpose:		permet de basculer, du mode demo au mode jeu
  -------------------------------------------------------*/
void switch_demo_to_game(void)
{
    cerr << "\t\tdone."<< endl;

    init_game();        
    cerr << "Starting:\trun_game()"<< endl;
    
    /*- reassigning the glut functions -*/
    glutDisplayFunc(run_game);
    glutIdleFunc(run_game);
    glutPassiveMotionFunc(motion);
}

static void reshape(int w, int h)
{
    glViewport(0, 0, (GLint)w, (GLint)h);
}

/*--------------------------------------------------------
    function: switch_light_on
    purpose:		allume ou eteint une lumire
  -------------------------------------------------------*/
void switch_light_on(GLenum which_one)
{
    if ( glIsEnabled(which_one) == GL_TRUE )
	    glDisable(which_one);
	else
	    glEnable(which_one);
}
						    
void Key( unsigned char key, int x, int y )
{
    switch(key) 
    {
          case 27:  /* Escape */
				    cerr << "\nExit requested by user." << endl;
				    time(&end);
				    elapsed = end - start;
			    
				    if ( elapsed != 0 )
					    cerr << "Estimated framerate: " << (float)frame/elapsed << endl;
					    
				    cerr << "\nShutting down.\n";
				    cerr << "---------------------------------------------" << endl;				    
				    exit(0);
				    break;
					 
		    case '0':
				    switch_light_on(GL_LIGHT0);
					break;
					 
		    case '1':
				    switch_light_on(GL_LIGHT1);
					break;
					  
		    case '2':
				    switch_light_on(GL_LIGHT2);
					break;
					   
		    case '3':
				    switch_light_on(GL_LIGHT3);
				    break;
					 
		    case '4':
					switch_light_on(GL_LIGHT4);
				    break;
					 
		    case '5':
				    switch_light_on(GL_LIGHT5);
				    break;
					 
		    case 't':
				    for(int i = 0; i < map_list.getLength(); i++)
				    {
					    if (map_list[i]->isEnable() != true)
						    map_list[i]->Enable();
					    else
						   map_list[i]->Disable();
				    }
	  				make_room();
				    make_bank();
				    break;
					   					   
            case ' ':
			    if (cam_mov == 1)
			    {
				    cam_mov = 0;
				    vertex v,p;
				    
				    v = object_list[0]->getPosition() ;
				    p = object_list[0]->getPivot() ;
				    
				    object_list[QUEUE]->setPosition(v.x ,-0.9, v.z);
				    object_list[QUEUE]->setPivot(0.0, 0.0, 1.0);
				}
				else
				    cam_mov = 1;
				    
			    break;
			    
			case '+':
				    force += 0.01;
				    break;
				    
			case '-':
				    if (force >= 0.01 )
					    force -= 0.01;
				    break;
			
		    case 's':
				    if (game_mode == 0 )
					    switch_demo_to_game();
				    break;
				    
		    case 'c':
				    if (players_turn == 0 )
					    players_turn =1;
					else
					    players_turn = 0;
				    break;
				    
			default:
				    return;
    }
}

void SpecialKey( int key, int x1, int y1 )
{
  float f = 0.2;
  
  if (cam_mov == 1)
  {
      switch(key)
      {
        case GLUT_KEY_UP:
            eye.Travel( 0, 0, f );
            break;
	    
        case GLUT_KEY_DOWN:
            eye.Travel( 0, 0, -f);
            break;
            
        case GLUT_KEY_LEFT:
            eye.Travel( -f, 0, 0 );
            break;
            
        case GLUT_KEY_RIGHT:
            eye.Travel( f, 0, 0 );
            break;
            
        case GLUT_KEY_PAGE_UP:
            eye.Travel( 0.0, f, 0.0 );
            break;
	    
        case GLUT_KEY_PAGE_DOWN:
            eye.Travel( 0.0, -f, 0.0 );
            break;
                    
        default:
            return;
      }
  }
  else
  {
      switch(key)
      {
        case GLUT_KEY_LEFT:
		    object_list[QUEUE]->Rotate(0.0, -2.0, 0);
            break;
            
        case GLUT_KEY_RIGHT:
		    object_list[QUEUE]->Rotate(0.0, 2.0, 0);
            break;
            
        default:
            return;
	    }
  }
  glutPostRedisplay();
}

void motion(int x, int y)
{
    if (cam_mov == 1 )
    	    eye.setRotation( (x-SCREEN_X_CENTER)*cam_rot_fact , (y-SCREEN_Y_CENTER)*cam_rot_fact, 0.0) ;
	else
	    object_list[QUEUE]->setRotation( 0.0, (x-SCREEN_X_CENTER)*queue_rot_fact , 0.0 ) ;
}

void mouse(int button, int state, int x, int y)
{
    if(button == GLUT_LEFT_BUTTON)
	    hit_ball(players_turn);
}


/*--------------------------------------------------------
    function: check_options
    purpose:		verifie les options de ligne de commande
  -------------------------------------------------------*/
void check_options( int a, char **c)
{
    int i = 0;
    
	while( i < a )
	{
   		i++;

	    if ( i >= a ) 
		    break;
	
	    if ( strncmp( c[i], "-width=",7) == 0 )
	    {
		    if ( ( atoi( c[i]+7 ) < 1024 )&&( atoi( c[i]+7) > 0 ) )
			    SCREEN_WIDTH = atoi( c[i]+7 );
			continue;
		}
		
	    if ( strncmp( c[i], "-height=",8) == 0 )
	    {
		    if ( ( atoi( c[i]+8) < 768 )&&( atoi( c[i]+8) > 0 ) )
			    SCREEN_HEIGHT = atoi( c[i]+8);
			continue;			    
		}
		
	    if ( strncmp( c[i], "-quality=",9) == 0 )
	    {
		    if ( strncmp( c[i]+9, "high",4) == 0 )
			    image_quality = HIGH_QUALITY ;
			    
		    if ( strncmp( c[i]+9, "normal",6) == 0 )
			    image_quality = NORMAL_QUALITY ;
			    
    		    if ( strncmp( c[i]+9, "low",3) == 0 )
			    image_quality = LOW_QUALITY ;
			continue;
		}
		
	    cerr << "unknown options" << endl;
		cerr << "Usage: " << c[0] << "-width= -height= -quality=[low normal high]" << endl;
		cerr << "-width sets the window's width" <<endl;
		cerr << "-height sets the window's height" <<endl;
		cerr << "-quality sets the rendering quality" <<endl;
		cerr << endl << "Example:" << endl;
		cerr << c[0] << " -width=640 -height=480 -quality=high" << endl;
		exit(1);
    }
    
	SCREEN_X_CENTER = SCREEN_WIDTH / 2;
	SCREEN_Y_CENTER = SCREEN_HEIGHT / 2;
    ASPECT_RATIO = (float)SCREEN_WIDTH/(float)SCREEN_HEIGHT;

    cam_rot_fact = 720.0 / SCREEN_WIDTH ;
    queue_rot_fact = 720.0 / SCREEN_WIDTH ;
}

int main(int argc, char *argv[])
{
    system("clear");
    cerr << "---------------------------------------------" << endl;    
	cerr << "|        <-[ Starting glBillard() ]->       |" << endl;
	cerr << "|                         Eneko Bilbao 1999 |" << endl;
    cerr << "---------------------------------------------" << endl;
    
	timer = 0;
    time_slice = 0.1 ;
    
    
    /* Camera init */
	eye.setPosition( -3.0, 1.5, 3.0);
    eye.setUpperLimit( -7.5, -3.0, -13.5);
    eye.setDownLimit(  7.5, 3.0, 13.5);
    
    check_options( argc, argv );
    
    if (image_quality == HIGH_QUALITY) 
	    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_ACCUM );
    else
	    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH );
    
    glutInitWindowPosition(100, 100);
    glutInitWindowSize(SCREEN_WIDTH,SCREEN_HEIGHT);
    glutCreateWindow("OpenGL Billard");
    
/*  GameMode does not seem to work... I couldn't get a valid screen mode :(
    if (argc > 1 && !strcmp(argv[1], "-fullscreen")) {
	    glutGameModeString("1152x864:16@72");
	    glutEnterGameMode();
	} else {
	    glutInitWindowPosition(100, 100);
	    glutInitWindowSize(SCREEN_WIDTH,SCREEN_HEIGHT);
	    glutCreateWindow("openGL Billard");
    }
*/
   	query_system();
    init_render_engine();
    load_textures();
    
	init_demo();
	
	glutDisplayFunc(run_demo);
    glutReshapeFunc(reshape);
	
    glutIdleFunc(run_demo);
    
    glutMouseFunc(mouse);
    glutKeyboardFunc(Key);
    glutSpecialFunc(SpecialKey);
    
    glutWarpPointer(SCREEN_X_CENTER,SCREEN_Y_CENTER);
    
    cerr << "Starting:\trun_demo()";
    time(&start);

    glutMainLoop();
    return 0;
}
