#include"mesh.h"


/*
  racine carr  bannir 
  voir les sries de Taylor ou du KNI  l'avenir
*/

void Mesh::computeFaceNormals( void )
{
	vertex p1, p2, p3;
	float l;
	
	for( int i=0 ; i < nb_faces ; i++) {
		
		p1 =  coords[skin[i].p1];
		p2 =  coords[skin[i].p2];
		p3 =  coords[skin[i].p3];
		
		skin[i].normal.x =  ( p2.z - p1.z ) * ( p1.y - p3.y ) - ( p2.y - p1.y ) * ( p1.z - p3.z ) ;
		skin[i].normal.y =  ( p2.x - p1.x ) * ( p1.z - p3.z ) - ( p2.z - p1.z ) * ( p1.x - p3.x ) ;
		skin[i].normal.z =  ( p2.x - p1.x ) * ( p1.y - p3.y ) - ( p2.y - p1.y ) * ( p1.x - p3.x ) ;

		l = sqrt( skin[i].normal.x * skin[i].normal.x + skin[i].normal.y * skin[i].normal.y + skin[i].normal.z * skin[i].normal.z );
		
		skin[i].normal.x /=  l ;
		skin[i].normal.y /=  l ;
		skin[i].normal.z /=  l ;
	}
}

float Mesh::getVertexLength( const int i)
{
	return(coords[i].l);
}
	
void Mesh::computeVertexNormals( void )
{
	for( int i=0 ; i < nb_vertex ; i++)
		coords[i].l = sqrt( coords[i].x * coords[i].x + coords[i].y * coords[i].y + coords[i].z * coords[i].z );
}

vertex& Mesh::getFaceNormal( int n )
{
	return( skin[n].normal );
}

float Mesh::getFaceNormalX( int n )
{
	return( skin[n].normal.x );
}

float Mesh::getFaceNormalY( int n )
{
	return( skin[n].normal.y );
}

float Mesh::getFaceNormalZ( int n )
{
	return( skin[n].normal.z );
}

void Mesh::setFaceSide( const int n, const char c )
{
	if ( n < nb_faces )
		skin[n].side = c ;
}

char Mesh::getFaceSide( const int n )
{
	return( skin[n].side ) ;
}

void Mesh::setFaceColor( int n, float r, float g, float b, float a)
{
	if ( n < nb_faces )
	{
		skin[n].color[0] = r ;
		skin[n].color[1] = g ;
		skin[n].color[2] = b ;
		skin[n].color[3] = a ;
    }
}

float *Mesh::getFaceColor( int n )
{
	return(skin[n].color);
}

int Mesh::compareFace( const face f1, const face f2)
{
	if ( f1.z > f2.z ) 
		return(-1);
	
	if ( f1.z < f2.z ) 
		return(1);
	
	return(0);
}

void Mesh::Sort( int d, int f, face *p )
{
	int i, j;
	face tmp;
	
	if ( d < f )
	{
		i = d + 1;
		while( (compareFace( *(p+i), *(p+d) ) == -1)&&( i < d) ) i++;
		
		j = f;
		
		while( compareFace( *(p+j), *(p+d) ) == 1 ) j--;
		
		while( i  <  j )
		{
			tmp = *( p + i );
			*( p + i ) = *( p + j );
			*( p + j ) = tmp;
			
			i++;
			while ( (compareFace( *(p+i), *(p+d) ) == -1)&&( i < d) ) i++;
			
			j--;
		
			while ( compareFace( *(p+j), *(p+d) ) == 1 ) j--;
		}
			tmp = *(p+d);
			*(p+d) = *(p+j);
			*(p+j) = tmp;
				
			Sort( d , j-1 , p );
			Sort( j+1 ,  f, p );
	}
	return ;
}

		
void Mesh::computeZFace( void )
{
	for( int i=0; i < nb_faces ; i++)
	{
		skin[i].z += coords[skin[i].p1].z ;
		skin[i].z += coords[skin[i].p2].z ;
		skin[i].z += coords[skin[i].p3].z ;
		skin[i].z /= 3;
	}
}
	
void Mesh::sortZFace( void )
{
	Sort( 0, nb_faces-1, skin );
}
		
void Mesh::setPivot( float x, float y, float z)
{
	pivot.x = x ; pivot.y = y ; pivot.z = z ; 
}

void Mesh::updatePivot( void)
{
	for( int i=0; i < nb_vertex ; i++)
	{
		pivot.x += coords[i].x ;
		pivot.y += coords[i].y ;
		pivot.z += coords[i].z ;
	}
	
	pivot.x /= nb_vertex ;
	pivot.y /= nb_vertex ;
	pivot.z /= nb_vertex ;
}

bool Mesh::setVertex( int n, float x, float y, float z)
{
	if ( n >= max_vertex ) 
		return( false );
	else
	{
		coords[n].x = x;
		coords[n].y = y;
		coords[n].z = z;
		nb_vertex++;
	}
	return( true );
}

void Mesh::addVertex( float x, float y, float z)
{
    coords[nb_vertex].x = x;
    coords[nb_vertex].y = y;
    coords[nb_vertex].z = z;
    nb_vertex++;
}
		
void Mesh::addFace( unsigned int d1, unsigned int d2, unsigned int d3)
{
    skin[nb_faces].p1 = d1;
    skin[nb_faces].p2 = d2;
    skin[nb_faces].p3 = d3;
    nb_faces++;
}

bool Mesh::setFace( int n, unsigned int d1, unsigned int d2, unsigned int d3)
{
	if ( n >= max_faces ) 
		return( false );
	else
	{
		skin[n].p1 = d1;
		skin[n].p2 = d2;
		skin[n].p3 = d3;
	}
	return( true );
}
	
vertex& Mesh::getVertex( int n )
{
	if ( n >= max_vertex ) 
		return( coords[0] );
	else
		return ( coords[n] );
}

vertex& Mesh::getCoords( int n )
{
	if ( n >= max_vertex ) 
		return( coords[0] );
	else
		return ( coords[n] );
}

void Mesh::Translate( float x, float y, float z )
{
	for( int i=0; i < nb_vertex ; i++)
	{
		coords[i].x += x;
		coords[i].y += y;
		coords[i].z += z;
	}
}

void Mesh::Translate( vertex& dist )
{
	for( int i=0; i < nb_vertex ; i++)
	{
		coords[i].x += dist.x;
		coords[i].y += dist.y;
		coords[i].z += dist.z;
	}
	updatePivot();
}

void Mesh::Scale( float x, float y, float z )
{
	for( int i=0; i < nb_vertex ; i++)
	{
		coords[i].x *= x;
		coords[i].y *= y;
		coords[i].z *= z;
	}
	
	updatePivot();
	computeFaceNormals();
}

void Mesh::Scale( vertex& dist )
{
	for( int i=0; i < nb_vertex ; i++)
	{
		coords[i].x *= dist.x;
		coords[i].y *= dist.y;
		coords[i].z *= dist.z;
	}
	updatePivot();
	computeFaceNormals();
}

/*
  updating the normals sucks
  should rotate them instead...
*/
void Mesh::RotateX( int a )
{
	float ac, as;
	float y ;
	
	a %= 360;
	if ( a < 0 )  a += 360 ;
	
	ac = trigo.COS(a);
	as = trigo.SIN(a);
	
	for( int i=0; i < nb_vertex ; i++)
	{
		y = coords[i].y;
		coords[i].y = coords[i].y * ac - coords[i].z * as ;
		coords[i].z = coords[i].z * ac + y * as ;
	}
	updatePivot();
	computeFaceNormals();
}

void Mesh::RotateY( int a )
{
	float ac, as;
	float x;
	
	a %= 360;
	if ( a < 0 )  a += 360 ;
	
	ac = trigo.COS(a);
	as = trigo.SIN(a);
	
	for( int i=0; i < nb_vertex ; i++)
	{
		x = coords[i].x;
		coords[i].x = coords[i].x * ac - coords[i].z * as ;
		coords[i].z = coords[i].z * ac + x * as ;
	}
	
	updatePivot();
	computeFaceNormals();
}

void Mesh::RotateZ( int a )
{
	float ac, as;
	float x ;
	
	a %= 360;
	if ( a < 0 )  a += 360 ;
	
	ac = trigo.COS(a);
	as = trigo.SIN(a);
	
	for( int i=0; i < nb_vertex ; i++)
	{
		x = coords[i].x;
		coords[i].x = coords[i].x * ac - coords[i].y * as ;
		coords[i].y = coords[i].y * ac + x * as;
	}
	updatePivot();
	computeFaceNormals();
}		

void Mesh::rotatePoint( int a, int b, int c, vertex& axe )
{
	float ac, as, bc, bs, cc, cs;
	float x, y, z;
	
	a %= 360;
	b %= 360;
	c %= 360;
	
	if ( a < 0 )  a += 360; 
	
	if ( b < 0 )  b += 360; 
	
	if ( c < 0 )  c += 360; 
	
	ac = trigo.COS(a);
	as = trigo.SIN(a);
	
	bc = trigo.COS(b);
	bs = trigo.SIN(b);

	cc = trigo.COS(c);
	cs = trigo.SIN(c);

	for( int i=0; i < nb_vertex ; i++)
	{
		x  = coords[i].x - axe.x ;
		y  = coords[i].y - axe.y ;
		z  = coords[i].z - axe.z ;
		// matrix better ?
		coords[i].x = axe.x + ( ( x * bc * cc ) + ( y * ( as * bs *cc - ac * cs) ) + ( z * ( ac * bs *cc + as *cs) ) );
		coords[i].y = axe.y + ( ( x * bc * cs ) + ( y * ( as * bs *cs + ac * cc) ) + ( z * ( ac * bs *cs - as *cc) ) );
		coords[i].z = axe.z + ( ( x * -bs) + ( y * as * bc ) + ( z * ac * bc ) );
	}
	computeFaceNormals();
}

void Mesh::rotatePivot( int a, int b, int c )
{
	float ac, as, bc, bs, cc, cs;
	float x, y, z;
	
	a %= 360;
	b %= 360;
	c %= 360;
	
	if ( a < 0 )  a += 360; 
	if ( b < 0 )  b += 360; 
	if ( c < 0 )  c += 360; 
	
	ac = trigo.COS(a);
	as = trigo.SIN(a);
	
	bc = trigo.COS(b);
	bs = trigo.SIN(b);

	cc = trigo.COS(c);
	cs = trigo.SIN(c);

	for( int i=0; i < nb_vertex ; i++)
	{
		x  = coords[i].x - pivot.x ;
		y  = coords[i].y - pivot.y ;
		z  = coords[i].z - pivot.z ;
		
		coords[i].x = pivot.x + ( ( x * bc * cc ) + ( y * ( as * bs *cc - ac * cs) ) + ( z * ( ac * bs *cc + as *cs) ) );
		coords[i].y = pivot.y + ( ( x * bc * cs ) + ( y * ( as * bs *cs + ac * cc) ) + ( z * ( ac * bs *cs - as *cc) ) );
		coords[i].z = pivot.z + ( ( x * -bs) + ( y * as * bc ) + ( z * ac * bc ) );
	}

/*	NEED TO FIX THIS, SINCE RECALCULATING IS TOO EXPENSIVE 
	
	cerr << "Before: " << endl;
	for( int i = 0 ; i < nb_faces ; i++ )
	{
		x = skin[i].normal.x + pivot.x;
		y = skin[i].normal.y + pivot.y;
		z = skin[i].normal.z + pivot.z;
		
		skin[i].normal.x = ( ( x * bc * cc ) + ( y * ( as * bs *cc - ac * cs) ) + ( z * ( ac * bs *cc + as *cs) ) ) ;
		skin[i].normal.y = ( ( x * bc * cs ) + ( y * ( as * bs *cs + ac * cc) ) + ( z * ( ac * bs *cs - as *cc) ) ) ;
		skin[i].normal.z = ( ( x * -bs) + ( y * as * bc ) + ( z * ac * bc ) ) ;
		
		cerr << "Face n: " << i << " x: " << skin[i].normal.x 
		   						 << " y: " << skin[i].normal.y 
								 << " z: " << skin[i].normal.z << endl;
	}
*/
	computeFaceNormals();
/*	
	cerr << "After: " << endl;
	for( int i = 0 ; i < nb_faces ; i++ )
		cerr << "Face n: " << i << " x: " << skin[i].normal.x 
		    << " y: " << skin[i].normal.y 
		    << " z: " << skin[i].normal.z << endl;
*/	
}

/*--------------------------------------------------------------------------*/
/*    MeshLoader                                                            */
/*--------------------------------------------------------------------------*/
/*  PURPOSE: loads a 3DSMAX ASCII export                                    */
/*--------------------------------------------------------------------------*/
void Mesh::loadMesh(char *filename)
{
  int nb_vertex, nb_faces, p1, p2, p3, i=0;
  float x, y, z;
  
  FILE *f;
  
  f = fopen( filename, "r" );
  
  if ( f == 0 ) {
    cerr << "File not found." << endl;
    return ;
  }
  
  fscanf( f, "%d\n", &nb_vertex );
  fscanf( f, "%d\n", &nb_faces );
  
  fscanf( f, "\n");
  
  Init(nb_vertex, nb_faces);
  
  while( i < nb_vertex ) {
    fscanf( f, "  %f %f %f\n",&x, &y, &z );
    addVertex( x, y, z );
    i++;
  }
  
  fscanf( f, "\n");
  
  i = 0;

  while( i < nb_faces ) {
    fscanf( f, " %d %d %d\n", &p1, &p2, &p3 );
	addFace( p1, p2, p3);
    
    if ( i % 2 == 0 )
      setFaceSide( i, 0);
    else
      setFaceSide( i, 1);
       
    i++;
  }
  
	for( i = 0 ; i < nb_vertex ; i++)
		setFaceColor( i, 0.0, 0.0, 1.0, 1.0 );

	updatePivot();
	computeFaceNormals();
	computeVertexNormals();
}

