/*---------------------------------------------------------------------------

															      Mesh v1.0
  Eneko Bilbao 1999
---------------------------------------------------------------------------*/

/* 
    Loads and handles 3ds Max ASE
	output files.
*/

/* STATUS
    -   transformations are all working.
    -		transformations around the pivot works.
*/

/* TODO
    -		remove the normal recalculation after a rotation,
	    rotate the normals instead (faster).
    -		add support for texture coordinates.
    -		prepare a hierarchical system.
    -		use matrix, to take advantage of hardware support.
    -   use triangle strips.
    
*/    

/* BUGS
    -	 rotation of normal vectors
*/

#ifndef _MESH_H_
#define _MESH_H_

#include"utility.h"
#include"vertex.h"

#include<math.h>
#include<stdlib.h>
#include<iostream.h>
#include<stdio.h>


class face{
	public:
		unsigned int  p1, p2, p3;
		vertex normal;
		char side;
		float z;
		float color[4];
};

class Mesh {

	private:
			char *name;
			
			vertex *coords;
			vertex *vertex_normals;
			
			face *skin;
			vertex *face_normals;
			
			vertex pivot;
			
			int nb_vertex;
			int max_vertex;
			
			int nb_faces;
			int max_faces;
			
			utility trigo;
			
			int compareFace( const face, const face );
			void Sort( int , int , face * )	;
					
	public:
			Mesh(void)
			{
				coords = 0;
				skin = 0;
				max_vertex = 0 ;
				max_faces = 0;
			}
			
			Mesh (const int v, const int s)
			{
				coords = new vertex[v];
				vertex_normals = new vertex[v];
				
				face_normals = new vertex[s];
				skin = new face[s];
				
				max_vertex = v;
				max_faces = s;
				
				nb_faces = 0;
				nb_vertex = 0;
			}
			
			~Mesh ( )
			{
				delete[] face_normals;
				delete[] vertex_normals;
				delete[] skin;
			}
			
			void Init( const int v, const int s) // v: vertex s: face
			{
				coords = new vertex[v];
				vertex_normals = new vertex[v];
				
				face_normals = new vertex[s];
				skin = new face[s];
						
				max_vertex = v;
				max_faces = s;
				
				nb_faces = 0;
				nb_vertex = 0;
			}
			
			inline float getX( const int n )
			{ return ( coords[n].x ); }
			
			inline float getY( const int n )
			{ return ( coords[n].y ); }
			
			inline float getZ( const int n )
			{ return ( coords[n].z ); }
			
			bool setX( const int n, const float x)
			{
				if ( n >= max_vertex )
					return( false );
				else
					coords[n].x = x;
				return ( true);
			}
			
			bool setY( const int n, const float y)
			{
				if ( n >= max_vertex )
					return( false );
				else
					coords[n].y = y;
				return ( true );
			}
			
			bool setZ( const int n, const float z)
			{
				if ( n >= max_vertex )
					return( false );
				else
					coords[n].z = z;
				return ( true);
			}
			
			inline int getNbPoints( void )
			{ return ( nb_vertex ); }
			
			inline int getNbFaces( void )
			{ return ( nb_faces ); }
			
			inline int getPointFromFace( const int f, const int c) // f = numero de la face et c la coordonne
			{
				switch( c )
				{
					case 0:
						return( skin[f].p1 );
						
					case 1:
						return( skin[f].p2 );

					case 2:
						return( skin[f].p3 );
				
					default: 
						cerr << " Error: invalid face number ! " << endl;
						return( -1 );
				}
			}
			
			void setPivot(float , float, float );
			vertex& getVertex( const int );
			void addVertex( const float, const float, const float);
			bool setVertex( const int , const float, const float, const float);
			
			void addFace( const unsigned int, const unsigned int, const unsigned int);
			bool setFace( const int , const unsigned int, const unsigned int, const unsigned int);
			
			void setFaceSide( const int , const char );
			char getFaceSide( const int);
			
			void setFaceColor( int , float, float, float, float );
			float *getFaceColor(int);
			
			float getFaceNormalX( const int );
			float getFaceNormalY( const int );
			float getFaceNormalZ( const int );
			
			void computeZFace( void );
			void sortZFace(void );
			
			void computeFaceNormals(void);
			vertex& getFaceNormal( const int );
			
			void computeVertexNormals( void );
			float getVertexLength( const int );
			
			void Update(void);
			void updatePivot(void);

			void RotateX( const int );
			void RotateY( const int );
			void RotateZ( const int );			
					
			void rotatePoint( const int, const int, const int, vertex& );
			void rotatePivot( int , int , int );
			
			void Translate( float, float, float );
			void Translate( vertex& );
			
			void Scale( float, float, float ) ;
			void Scale( vertex& ) ;
			vertex& getCoords( const int );
			
			void setTexture(int);
			void getTexture(void);
			
			void loadMesh(char*);
			
};
#endif /*---------------------------------------------*/





